% cv. 6 ; na 29.3.2017 (St) ; Jan Hric
% v. 1.3

% Obsah: cv6

% v navaznosti na minule: 
% - operatory a vyrokove formule
%   - pomoci minuleho src: eval 
%   - CV: propagace non/1 dolu
% - tok dat: asynchronni odebirani ze seznamu  
% 	naCasti/3
% - heapsort
%   - rozbor
%   - vyroba tvaru haldy, 2x
%   - CV: heap2srt/2 a slucHaldy/3
%   DCv: zhalduj/2
% - vlozene operace
%   - idiom: generuj a testuj



%%%%%%%%%% z minula: explicitni vysledky:
%
% compare/3: na prednasce
% ?- compare( < , 3, 5). % 1.arg
%
% - idiom: vysledky jako data
% podobne pouziti: v Prologu pouzivane uspeje/neuspeje jsme v eval/3 
%   nahradili explicitnim true/false (a eval/3 vzdy uspeje)
% podobne pouziti: vysledky misto backtrackingu vracime jako seznam reseni


%%%%%%%%
:- op(210, xfy, [and]).  % def. operatoru and/2
:- op(220, xfy, or).  % def. operatoru 
:- op(200, fy, non).  % def. operatoru 
% - non/1 umozni display(non non a)).

%%%%%%%%%%%%%%%%%%%%%%%
% CV: propagace non/1 dolu na non, and, or, a p/1 true/0, false/0
% dve urovne rekurze ...
propagNon(VF,VF) :- true.
propagNon(VF1 and VF2, VF1 and VF2) :- true.
propagNon(VF1 or VF2, VF1 or VF2) :- true.
propagNon(non VF, non VF) :- true.
propagNon(non non VF, non non VF) :- true.
propagNon(non (_VF1 and _VF2), _V) :- true. % 
propagNon(non (_VF1 or  _VF2), _V) :- true. % 
% ... rozbor ppad u non/1

elementarniVF(VF):- 
  VF = p(_) ; 
  VF = true ;
  VF = false .

tg1(V) :- propagNon( non(p(a) and non(p(b) or non(true))), V).

%%%%%%%%%%%%%%%%%%
% rozdel na casti: dualni k rozdilovym seznamum
%
% naCasti(+[Delky],+[Prvky],-[[Casti]]) :- ze seznamu Prvky ubere casti pozadovane Delky
%   Chceme: Delky se musi dat odebrat, prvk musi byt dost.
% priklad na tok dat

naCasti([],_,[]).
% naCasti([_|_],[],[]).  % jina okrajova podminka
naCasti([D|Ds],Xs,[V|Vs]):-
  uberCast1(D,Xs,V,Xs1),
  !, 	% mame reseni
  naCasti(Ds,Xs1,Vs).
% varianta: v dalsim argumentu zbyle prvky 

uberCast(D,Xs,V,Xs1) :-
  length(V,D),  % vyrobi seznam volnych promennych 
  append(V,Xs1,Xs).	% naplni ho
% length/2 nejdriv je ponekud efektivnejsi nez (opakovane) kontrolovani odebrane delky

% varianta, syntakticka 
uberCast1(0,Xs,[],Xs).	% to same, explicitni rekurze
uberCast1(D,[X|Xs],[X|V],Xs1) :-
  D>0, D1 is D-1,
  uberCast1(D1,Xs,V,Xs1).

% volani:
% ?- naCasti([1,2,3],[a,b,c,d,e,f,g,h,i],V]).
%  V=[[a],[b,c],[d,e,f]]
% ! tok dat:
% naCasti([1,2,3],Xs,[V1,V2,V3]) :-
%   uberCast(1,Xs,V1,Xs1), uberCast(2,Xs1,V2,Xs2), uberCast(3,Xs2,V3,_Xs3).
% vstupni seznam Xs se postupne zkracuje na Xs1, Xs2, ...


%%%%%%%%%%%%%%%%%%%%%%%%%%
% heapSort: rozbor:
% heapsort (s binarni haldou)
% - postav tvar heapu, zhalduj, vybirej
%   - vyuziva rozdel/3
%   - stavba: prvek pro vrchol utrhnu hned
% - stavba: doleva doprava
% - hned stavet haldu (rozdel a panuj)
%   - minimum do korene a zbytek rozdelit
%   - stavba v O(n log n)
%   - klasika je postavit 2 subhaldy a potom 1 prvek vlozit, O(n) asi
% - kapitansky krok: hloubka je porad logaritmicka
%   - (ani) pri vybirani nemusim udrzovat vyvazenou haldu
%     - posledni halda je neuplna 
%   - vybirani v O(n log n)


%%%%%%%%%%%%%%%%%
% heapsort
% CV: slucovani hald

heapsort(Ls,Vs):-
  mkHaldaKK(0,Ls,[],nil,Neheap), % vyrobi tvar haldy
  zhalduj(Neheap, Heap), % zajisti splneni haldovych podminek
  halda2srt(Heap,Vs).

%%%%%%%%%%
% CV , prevod haldy na usporadany seznam
halda2srt(nil,[]).
halda2srt(t(_L,_X,_P), Vs):- 
  true,
  halda2srt(_,Vs).

%%%%%%%%%  CV
% slucHaldy(L,P,V) :- slouci haldy L a P do V
slucHaldy(nil, P,   P).
slucHaldy(L  , nil, L).  
  % :- L \= nil. % uprava pro neprekryvani podminek
slucHaldy(t(_LL,LX,_LP), t(_PL,PX,_PP), V) :-
  LX =< PX ->	V = t(_,_,_),
		slucHaldy(L, P, VV)
           ;	V = t(_,_,_),
		slucHaldy(L, P, VV) .
% - koncove podminky se prekryvaji ( pro slucHaldy(nil,nil,nil) ), ale jsou deklarativne spravne

% varianta: mame L a P vcelku, nejvyssi uroven struktury se nebuduje znova
% slucHaldy(L,P,V) :- 
%    L = t(LL,LX,LP), P = t(PL,PX,PP), 
%    ( LX <= PX  -> V = t(VL,LX,P), % P vcelku
%		    slucHaldy(_,_,VL)
%    ; ... )	% symetricky 
% - zavorky okolo if->then;else potrebujeme

%%%%%%%%%%%%%%%%%%
% - kapitansky krok: postupne zvetsovani hranice, exponencialne (tj. nasobne v kazdem kroku)
% - pro heapsort neprodukuje vyvazenou haldu, ale log. hloubka staci
% - cte data ve vstupnim poradi 

%%%%%%%%%%%%%%%%%%
% mkHaldaKK(+D,+Ls,-Ls0,+H1,-Heap) :- z Ls vyrabi haldu Heap hloubky D kapitanskym krokem, 
%	H1 je prubezne vyrobena, Ls0 jsou zbyle prvky
mkHaldaKK(_D,[],[],Heap,Heap).
mkHaldaKK(D ,[X|Ls1],Ls0,Heap1,Heap0) :-
  mkHaldaD(D,Ls1,Ls2,Heap2),
  D1 is D+1,
  mkHaldaKK(D1,Ls2,Ls0,t(Heap1,X,Heap2),Heap0).

% mkHaldaD(+D,+Ls1,-Ls0,-Heap) :- vyrobi haldu hloubky D, zbyle prvky v Ls0
mkHaldaD(0,Ls,Ls,nil) :- !. % rez kvuli druhe koncove podmince
mkHaldaD(_D,[],[],nil).     % dosly prvky
mkHaldaD(D,[X|Ls1],Ls0,t(H1,X,H2)):-
  D>0, D1 is D-1,
  mkHaldaD(D1,Ls1,Ls2,H1),
  mkHaldaD(D1,Ls2,Ls0,H2).

tg2(Xs-V):- mkHaldaD(3,[e,d,c,b,a],Xs,V).
tg3(V):- mkHaldaKK(0,[e,d,c,b,a],[],nil,V).

/* mkHaldaD		mkHaldaKK
    e			    b
   / \			   / \
  d   a			  d   a
 / \			 / \
c   b			e   c

*/



%%%%%%%%%%%%%%
% vlozene operace:
% vlozeni(Xs, Strom, V) :- doplnit mezi cifry aritmeticke operace +, -, *, / 
%	a spojeni cifer do cisla, aby vysel pozadovany vysledek.
vlozeni(Xs,Xs,V) :- sezNaCislo(Xs,0,V).
  % vlozeni(Xs,cislo(Xs),V) :- ...  % pokud chceme odlisovat cisla tagem
vlozeni(Xs,Strom,V):-
  Strom = op(A1,Op,A2), % var s ops1/5: zakomentovat
  append(Xs1,Xs2,Xs), Xs1\=[], Xs2\=[],
  vlozeni(Xs1,A1,V1),
  vlozeni(Xs2,A2,V2),
  ops(Op,V1,V2,V).	% povolene operace, vcetne semantiky (4.arg) 
  % ops1(_Op,V1-A1,V2-A2,V,Strom). % pridani vystupni stromove syntaxe

% vypocet hodnoty z cifer (desitkove) pozicni soustavy
sezNaCislo([],V,V).
sezNaCislo([X|Xs],V1,V0) :-
  V2 is V1*10+X,
  sezNaCislo(Xs,V2,V0).

ops(+,V1,V2,V) :- V is V1+V2.
ops(*,V1,V2,V) :- V is V1*V2.
ops(-,V1,V2,V) :- V is V1-V2, V >= 0.
ops(/,V1,V2,V) :- V2>0, V is V1 // V2, V1 mod V2 =:= 0. % vcetne testu

ops1(+,V1-A1,V2-A2,V,A1+A2) :- V is V1+V2.
ops1(*,V1-A1,V2-A2,V,A1*A2) :- V is V1*V2.
ops1(-,V1-A1,V2-A2,V,A1-A2) :- V is V1-V2, V >= 0.
ops1(/,V1-A1,V2-A2,V,A1/A2) :- V2>0, V is V1 // V2, V1 mod V2 =:= 0. % vcetne testu

% na rozmysleni: prevest na seznam vysledku


%%%%%%%%%%
% DCv: zhalduj(+Neheap, -Heap) :- vytvorena Heap zachova tvar Neheap, 
%	ale zajisti haldovou podminku (predek < potomek)

/* %%%%%%%%%% 
  dalsi:
  - na asynchronni odbirani vstupu:
    - minimovy strom, z ADS1
 	Vstup a_1 az a_n. Stavime binarni strom. V koreni je minimum posloupnosti, 
 	necht je to a_j. Pak levy podstrom je minimovy strom z a_1 az a_(j-1) a pravy 
	podstrom minimovy strom z a_(j+1) az a_n. Konstrukce v O(n), klasicky. Urcen jednoznacne.
*/ 

%%%%%%%%%%%%%%%%%%%%
% souhrn: co jste si vyzkouseli (nebo meli zvladnout):
% - (rozdilove seznamy a) asynchronni odebirani vstupu, predavani zbytku stavem
% - zexplicitneni dat
% - zpracovani termovych (tj. stromovych) struktur (opet)
% - generuj a testuj



%%%%%%%%%%%%%


%%%%%%%%%%%%%%%
% stromy, binarni: funktory: t/3, nil/0
% proc. prohledani/2, vyda casy otevreni a uzavreni vrchol, v seznamu

/*  * /
/ *
*/

%%%%%%%%%%%%%%%%%%%%
% souhrn: co jste si vyzkouseli:
% -
% -
% -
% -
% -
% -
% -
% -

%%%%%%%%%%%%%%%%%%%%%%%%%%%
% runtime
check(Tested,=,Pattern) :- 
  Tested = Pattern -> write(passed),nl
                   ;  write(error),nl, 
		      write(value__(Tested)),nl, 
		      write(expected(Pattern)),nl.

%----------------

% k.
